// $Id: bignum.h 763 2010-01-28 23:15:33Z mag $
#ifndef BIGNUM_H_
#define BIGNUM_H_

#include <boost/random/mersenne_twister.hpp>

template <uint8_t N>
class BigNum{

    uint8_t _n[3*N];

    static mt19937 rng;
    static const uint8_t M = 3*N;

public:

    BigNum();
    BigNum(uint8_t l);
    BigNum(const BigNum<N>& bn);

    BigNum<N> random() const;

    operator bool () const;
    operator double () const;

    BigNum& operator += (const BigNum<N>& bn);
    BigNum& operator -= (const BigNum<N>& bn);

    bool operator < (const BigNum<N>& bn) const;
    bool operator <= (const BigNum<N>& bn) const;
    bool operator == (const BigNum<N>& bn) const;
    bool operator != (const BigNum<N>& bn) const;

    static void seed(uint32_t s) { rng.seed(s); }

};

template <uint8_t N>
mt19937
BigNum<N>::rng;

template <uint8_t N>
inline
BigNum<N>::BigNum()
{ for (uint8_t i = 0; i < M; i++) _n[i] = 0U; }

template <uint8_t N>
BigNum<N>::BigNum(uint8_t l)
{

    l *= N;

    register uint8_t b = l/8;

    for (uint8_t i = 0; i < M; i++)
        if (i == b)
            _n[i] = 1U << l%8;
        else
            _n[i] = 0U;

}

template <uint8_t N>
inline
BigNum<N>::BigNum(const BigNum<N>& bn)
{ for (uint8_t i = 0; i < M; i++) _n[i] = bn._n[i]; }

template <uint8_t N>
BigNum<N>
BigNum<N>::random() const
{

    static const uint8_t M2 = 2*M;

    BigNum<N> bn;

    if (*this)
    {

        uint8_t b[M2];
        uint8_t i, j, k;

        for (i = 0; i < M2; i++)
            b[i] = 0U;

        uint8_t n = 4U;
        uint32_t random = rng();

        for (i = 0; i < M; i++)
        {

            uint16_t carry = 0U;

            for (j = 0, k = i; j < M; j++, k++)
            {
                register uint8_t bk = b[k];
                uint16_t prod = static_cast<uint16_t>(random & 0xff)*_n[j] + carry;
                b[k] += static_cast<uint8_t>(prod & 0xff);
                carry = (prod >> 8) + (b[k] < bk);
            }

            for ( ; k < M2; k++)
            {
                register uint8_t bk = b[k];
                b[k] += static_cast<uint8_t>(carry);
                carry = b[k] < bk;
                if (carry == 0)
                    break;
            }

            if (--n == 0)
            {
                n = 4U;
                random = rng();
            }
            else
                random >>= 8;

        }

        for (i = 0, j = M; i < M; i++, j++)
           bn._n[i] = b[j];

        assert(bn < *this);

    }
    else
    {

        uint8_t n = 4U;
        uint32_t random = rng();

        for (uint8_t i = 0; i < M; i++)
        {

            bn._n[i] = static_cast<uint8_t>(random & 0xff);

            if (--n == 0)
            {
                n = 4U;
                random = rng();
            }
            else
                random >>= 8;

        }

    }

    return bn;

}

template <uint8_t N>
BigNum<N>::operator bool () const
{ 
    
    for (uint8_t i = 0; i < M; i++)
        if (_n[i])
            return true;

  return false;

}

template <uint8_t N>
BigNum<N>::operator double () const
{

    static const double bi = 256.0;
    static const double b = pow(2.0,-24*N);

    double v = 0.0;
    double p = 1.0;

    for (uint8_t i = 0; i < M; i++) {
        v += _n[i]*p;
        p *= bi;
    }

    return (v != 0.0) ? v*b : 1.0;

}

template <uint8_t N>
BigNum<N>&
BigNum<N>::operator += (const BigNum<N>& bn)
{

    bool carry = false;

    for (uint8_t i = 0; i < M; i++) {

        register uint8_t ni = _n[i];

        _n[i] += bn._n[i] + carry;

        carry = (_n[i] < ni) || (carry && bn._n[i] == 0xff);

    }

    return *this;

}

template <uint8_t N>
BigNum<N>&
BigNum<N>::operator -= (const BigNum<N>& bn)
{

    if (bn) {

        uint8_t i;
        uint8_t n[M];

        for (i = 0; i < M; i++)
            n[i] = ~bn._n[i];

        i = 0;
        bool carry = n[0] == 0xff;

        n[0]++;

        while (carry && ++i < M) {
            carry = n[i] == 0xff;
            n[i]++;
        }

        carry = false;

        for (i = 0; i < M; i++) {

            register uint8_t ni = _n[i];

            _n[i] += n[i] + carry;

            carry = (_n[i] < ni) || (carry && n[i] == 0xff);

        }

    }

    return *this;

}

template <uint8_t N>
bool
BigNum<N>::operator < (const BigNum<N>& bn) const
{ 
    
    for (uint8_t i = M-1; i > 0; i--)
        if (_n[i] != bn._n[i])
            return _n[i] < bn._n[i];

    return _n[0] < bn._n[0];

}

template <uint8_t N>
bool
BigNum<N>::operator <= (const BigNum<N>& bn) const
{ 
    
    for (uint8_t i = M-1; i > 0; i--)
        if (_n[i] != bn._n[i])
            return _n[i] < bn._n[i];
  
    return _n[0] <= bn._n[0];

}

template <uint8_t N>
bool
BigNum<N>::operator == (const BigNum<N>& bn) const
{ 
    
    for (uint8_t i = 0; i < M; i++)
        if (_n[i] != bn._n[i])
            return false;

    return true;

}

template <uint8_t N>
bool
BigNum<N>::operator != (const BigNum<N>& bn) const
{ 
    
    for (uint8_t i = 0; i < M; i++)
        if (_n[i] != bn._n[i])
            return true;

    return false;

}

#endif /*BIGNUM_H_*/
