// $Id: generator.h 763 2010-01-28 23:15:33Z mag $
#ifndef GENERATOR_H_
#define GENERATOR_H_

#include <cmath>
#include <fstream>
#include <iostream>

#include <boost/progress.hpp>

#include "tree.h"
#include "sample.h"
#include "counter.h"

template <uint8_t N>
class Generator {

    bool _p;
    double _v;

    Tree<N> _t;
    Counter<N> _c;

    unsigned int _i;

    boost::progress_display _h;

public:

    class Iterator: public Tree<N>::Iterator {

        Iterator(const typename Tree<N>::Iterator& i);

        friend class Generator;

    public:

        Iterator();

        bool operator != (const Iterator& i) const;

    };

    Generator(bool p);

    double volume() const;

    Iterator end() const;
    Iterator begin() const;

    bool iterate();

    void output(double t,
                const string& filename) const;

};

template <uint8_t N>
inline
Generator<N>::Generator(bool p) :
    _p(p), _v(0.0), _i(0UL), _h(100,cerr) {}

template <uint8_t N>
inline double
Generator<N>::volume() const
{ return _v; }

template <uint8_t N>
inline typename Generator<N>::Iterator
Generator<N>::begin() const
{ return Iterator(_t.begin()); }

template <uint8_t N>
inline typename Generator<N>::Iterator
Generator<N>::end() const
{ return Iterator(_t.end()); }

template <uint8_t N>
inline
Generator<N>::Iterator::Iterator() {}

template <uint8_t N>
inline
Generator<N>::Iterator::Iterator(const typename Tree<N>::Iterator& i) :
    Tree<N>::Iterator(i) {}

template <uint8_t N>
inline bool
Generator<N>::Iterator::operator != (const Iterator& i) const
{ const typename Tree<N>::Iterator& it =
    reinterpret_cast<const typename Tree<N>::Iterator&>(i);
  return it != *this; }

template <uint8_t N>
bool
Generator<N>::iterate()
{

    static uint32_t size0 = 0UL;

    if (!_t.empty()) {

        Sample<N> s(_t.generate());

        assert(_t.valid());

        if (s.valid())
        {

            _t.update(Sample<N>::push(s));

            assert(_t.valid());

            if (_p) {

                _c.reset(true);

                while (!_c.finished()) {

                    Sample<N> snew(s);

                    snew += _c.sample_offset();

                    if (!_t.update(Sample<N>::push(snew)))
                        Sample<N>::pop();

                    assert(_t.valid());

                    _c++;

                }

            }

        }

        _v = _t.volume();

        uint32_t size1 = static_cast<uint32_t>(100.0*_v);

        if (size1 > size0) {
            _h += size1 - size0;
            size0 = size1;
        }

#if defined(MEMORY_STATISTICS)
        if ((++_i)%1000 == 0)
            Tree<N>::memory_stats();
#else
        ++_i;
#endif

        return true;

    }

    _v = 1.0;

    if (_h.count() < 100)
        _h += 100 - _h.count();

#if defined(MEMORY_STATISTICS)
    Tree<N>::memory_stats();
#endif

    return false;

}

template <uint8_t N>
void
Generator<N>::output(double t,
                     const string& filename) const
{

    Cube<N> cube;
    vector< Sample<N> > samples;

    for (uint32_t i = 0; i < Sample<N>::getNumSamples(); ++i)
    {
       const Sample<N>& sample = Sample<N>::getSample(i);
       if (cube.contains(sample))
          samples.push_back(sample);
    }

    if (_h.count() < 100)
        cerr << endl;

    cerr << samples.size() << " samples; ";

    uint32_t hours = static_cast<uint32_t>(t/3600.0); t -= 3600.0*hours;
    uint32_t minutes = static_cast<uint32_t>(t/60.0); t -= 60.0*minutes;

    if (hours > 0) cerr << hours << 'h';
    if (minutes > 0) cerr << minutes << 'm';

    cerr.setf(ios::fixed);
    cerr << setprecision(3) << t << "s; " << samples.size()/t << " samples/s\n";
    cerr << _t.get_terminals() << " unresolved leaf nodes\n";
    cerr << _i << " iterations\n";

    if (filename != "") {

        ofstream out(filename.c_str(),ofstream::binary);

        if (!out) {
            cerr << "Unable to open output file " << filename << endl;
            return;
        }

        uint32_t s = static_cast<uint32_t>(samples.size());
        out.write((const char*)&s,sizeof(uint32_t));

        float x = Sample<N>::getRadius();
        out.write((const char*)&x,sizeof(float));

        for (uint32_t i = 0; i < samples.size(); i++)
        {
            const Sample<N>& s = samples[i];
            for (uint8_t j = 0; j < N; j++) {
                float x = s.get(j);
                out.write((const char*)&x,sizeof(float));
            }
        }

        out.close();

    }

}

#endif /*GENERATOR_H_*/
