// $Id: poisson.cpp 763 2010-01-28 23:15:33Z mag $
#include <cmath>
#include <ctime>
#include <string>
#include <sstream>
#include <iomanip>

#include <boost/timer.hpp>
#include <boost/program_options/parsers.hpp>
#include <boost/program_options/cmdline.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/options_description.hpp>

#include "generator.h"

#if !defined(NDIM)
#  error "Dimension not defined"
#endif

#if NDIM == 0
#  error "Dimension must be greater than 0"
#elif NDIM > 4
#  error "Dimension must not be greater than 4"
#endif

#if defined(OPENGL)
#  if NDIM == 2 || NDIM == 3

#    include <GL/glut.h>

static uint16_t resx;
static uint16_t resy;
static double ti = 0.0;
static double dt = 0.0;
static bool interactive = false;
#    if NDIM == 2
static GLuint dlist;
static GLuint flist;
static bool fillvis = false;
static bool diskvis = false;
static bool voxelvis = false;
#    elif NDIM == 3
static GLuint slist;
static GLdouble xln = 0.0;
static GLdouble yln = 0.0;
static GLdouble cam = 1.8;
static GLdouble mphi = 0.445;
static GLdouble mtheta = -0.688296;
static GLdouble cphi = M_PI/4.0;
static GLdouble ctheta = -M_PI/4.0;
static GLdouble eye[] = { 0.0, 0.0, cam };
static GLdouble upv[] = { 0.0, 1.0, 0.0 };
static GLdouble clp[] = { -0.5, -0.5*sqrt(2.0), -0.5, 0.5*sqrt(0.75) };
static bool clpvis = false;
static bool sphvis = false;
static bool clipflag = false;
static bool distflag = false;
static bool motionflag = false;
static GLint lastx = 0;
static GLint lasty = 0;
#    endif

#  endif
#endif

using namespace boost;
namespace po = program_options;

static float rad = 0.0f;
static uint32_t num = 0UL;
static bool periodic = false;

static string filename = "";


static timer poisson_timer;

static Generator<NDIM>* gen = 0;

#if defined(OPENGL)
#if NDIM == 2 || NDIM == 3

static void
parse_w(uint16_t& resx,
        uint16_t& resy,
        const string& spec)
{

    char ch;
    istringstream iss(spec);

    iss >> resx;

    if (iss.fail()) goto w_error;

    iss >> ch;

    if (iss.fail() || ch != ',') goto w_error;

    iss >> resy;

    if (iss.fail()) goto w_error;

    return;

w_error:
    std::cerr << "Could not read window resolution: " + spec << std::endl;
    exit(EXIT_FAILURE);

}

static void
display()
{

#if NDIM == 2
    glClear(GL_COLOR_BUFFER_BIT);
#elif NDIM == 3
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
#endif

    Sample<NDIM>::Iterator i;

    double volume = gen->volume();

#if NDIM == 2

    static Counter<NDIM> count;

    if (voxelvis) {

        Generator<NDIM>::Iterator i;

        for (i = gen->begin(); i != gen->end(); i++) {

            const Interval& ix = (*i).get(0);
            const Interval& iy = (*i).get(1);

            glColor3f(0.5f,0.0f,0.5f);
            glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

            glRectf(static_cast<float>(ix.min()),static_cast<float>(iy.min()),
                    static_cast<float>(ix.max()),static_cast<float>(iy.max()));

            glColor3f(1.0f,0.0f,1.0f);
            glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

            glRectf(static_cast<float>(ix.min()),static_cast<float>(iy.min()),
                    static_cast<float>(ix.max()),static_cast<float>(iy.max()));

        }

    }

    if (fillvis) {

        glColor3f(0.5f,0.5f,0.5f);

        for (i = Sample<NDIM>::begin(periodic); i != Sample<NDIM>::end(); i++) {
            glPushMatrix();
            glTranslatef(i->get(0),i->get(1),0.0f);
            glCallList(flist);
            glPopMatrix();
        }

    }

    glColor3f(0.0f,0.0f,0.0f);

    glBegin(GL_POINTS);

    for (i = Sample<NDIM>::begin(periodic); i != Sample<NDIM>::end(); i++)
        glVertex2f(i->get(0),i->get(1));

    glEnd();

    if (diskvis) {

        for (i = Sample<NDIM>::begin(periodic); i != Sample<NDIM>::end(); i++) {
            glPushMatrix();
            glTranslatef(i->get(0),i->get(1),0.0f);
            glCallList(dlist);
            glPopMatrix();
        }

    }

#elif NDIM == 3

    glLoadIdentity();

    glPushMatrix();

    gluLookAt(eye[0],eye[1],eye[2],0.0f,0.0f,0.0f,upv[0],upv[1],upv[2]);

    if (clpvis) {
        glClipPlane(GL_CLIP_PLANE0,clp);
        glEnable(GL_CLIP_PLANE0);
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
    }

    glColor3f(1.0f,0.0f,0.0f);
    glutWireCube(1.0);

    glTranslatef(-0.5f,-0.5f,-0.5f);

    if (sphvis) {

        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);

        for (i = Sample<NDIM>::begin(); i != Sample<NDIM>::end(); i++) {
            glPushMatrix();
            glTranslatef(i->get(0),i->get(1),i->get(2));
            glCallList(slist);
            glPopMatrix();
        }

        glDisable(GL_LIGHTING);
        glDisable(GL_LIGHT0);

    }
    else {

        glColor3f(0.0f,0.0f,0.0f);

        glBegin(GL_POINTS);

        for (i = Sample<NDIM>::begin(); i != Sample<NDIM>::end(); i++)
            glVertex3f(i->get(0),i->get(1),i->get(2));

        glEnd();

    }

    if (clpvis) {
        glDisable(GL_CLIP_PLANE0);
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
    }

    glPopMatrix();

#endif

    if (volume < 1.0) {

        glColor3f(1.0f,0.0f,0.0f);
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

#if NDIM == 2
        if (resx > resy) {
            double margin = 0.5*(resx - resy)/resy;
            glRectd(-margin,0.0,(1.0+margin)*volume,0.01);
        }
        else {
            double margin = 0.5*(resy - resx)/resx;
            glRectd(0.0,-margin,volume,0.01*(1.0+margin));
        }
#else
        glTranslatef(0.0f,0.0f,-static_cast<float>(cam));
        glRectd(-xln,-yln,xln*(2.0*volume - 1.0),-0.98*yln);
#endif

    }

    glutSwapBuffers();

    ti = poisson_timer.elapsed();

}

static void
reshape(int w, int h)
{

    glViewport(0,0,(GLsizei)(resx = w),
                   (GLsizei)(resy = h));

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

#if NDIM == 2

    if (resx > resy) {
        double margin = 0.5*(resx - resy)/resy;
        gluOrtho2D(-margin,1.0+margin,0.0,1.0);
    }
    else {
        double margin = 0.5*(resy - resx)/resx;
        gluOrtho2D(0.0,1.0,-margin,1.0+margin);
    }

#elif NDIM == 3

#define FOV 60.0

    static double l = sqrt(0.75);

    double aspect = static_cast<float>(w)/h;

    yln = tanf(static_cast<float>(FOV*M_PI/360.0))*cam;
    xln = aspect*yln;

    gluPerspective(FOV,aspect,cam - l - rad,cam + l + rad);

#endif

    glMatrixMode(GL_MODELVIEW);

}

static void
keyboard(unsigned char key, int x, int y)
{

#if NDIM == 2
    if (key == 0x64) {
        diskvis = !diskvis;
        fillvis = false;
        glutPostRedisplay();
    }
    else if (key == 0x66) {
        fillvis = !fillvis;
        diskvis = false;
        glutPostRedisplay();
    }
    else if (key == 0x76) {
        voxelvis = !voxelvis;
        glutPostRedisplay();
    }
#elif NDIM == 3
    if (key == 0x73) {
        sphvis = !sphvis;
        glutPostRedisplay();
    }
    else if (key == 0x63) {
        clpvis = !clpvis;
        glutPostRedisplay();
    }
#endif
    else if (key == 0x1b) {
        gen->output(dt,filename);
        exit(EXIT_SUCCESS);
    }

}

#if NDIM == 3

static void
camera_angles(GLdouble theta,
              GLdouble phi)
{

    GLdouble cy = cos(theta);
    GLdouble sy = sin(theta);
    GLdouble cz = cos(phi);
    GLdouble sz = sin(phi);

    eye[0] = cam*cy*cz;
    eye[1] = cam*sz;
    eye[2] = -cam*sy*cz;

    upv[0] = -cy*sz;
    upv[1] = cz;
    upv[2] = sy*sz;

    if (upv[1] < 0.0f) {
        upv[0] = -upv[0];
        upv[1] = -upv[1];
        upv[2] = -upv[2];
    }

}

static void
clip_angles(GLdouble theta,
            GLdouble phi)
{

    GLdouble cy = cos(theta);
    GLdouble sy = sin(theta);
    GLdouble cz = cos(phi);
    GLdouble sz = sin(phi);

    clp[0] = -cy*cz;
    clp[1] = -sz;
    clp[2] = sy*cz;

}

static void
mouse(int button,
      int state,
      int x, int y)
{

    if (button == GLUT_LEFT_BUTTON) {

        if (state == GLUT_DOWN) {
            lastx = x;
            lasty = y;
            motionflag = true;
        }
        else
            motionflag = false;

    }
    else if (clpvis) {

        if (button == GLUT_RIGHT_BUTTON) {

            if (state == GLUT_DOWN) {
                lastx = x;
                lasty = y;
                clipflag = true;
            }
            else
                clipflag = false;

        }
        else {

            if (state == GLUT_DOWN) {
                lastx = x;
                lasty = y;
                distflag = true;
            }
            else
                distflag = false;

        }

    }

}

static void
motion(int x, int y)
{

    if (motionflag || clipflag  || distflag) {

        GLdouble dx = (static_cast<GLdouble>(x) - lastx)/resx;
        GLdouble dy = (static_cast<GLdouble>(y) - lasty)/resy;

        if (motionflag) {

            mtheta -= 2.0*dx;
            mphi += 2.0*dy;

            camera_angles(mtheta,mphi);

        }
        else if (clipflag) {

            ctheta += 2.0*dx;
            cphi -= 2.0*dy;

            clip_angles(ctheta,cphi);

        }
        else {

            static GLdouble l = sqrt(0.75);

            clp[3] -= dy;

            clp[3] = min(clp[3],+l + rad);
            clp[3] = max(clp[3],-l - rad);

        }

        lastx = x;
        lasty = y;

        glutPostRedisplay();

    }

}

#endif

static void
iterate()
{

#define DELAY 0.16

    static double t0 = 0.0;

    if (t0 == 0.0)
        t0 = poisson_timer.elapsed();

    if (gen->iterate()) {

        dt = poisson_timer.elapsed() - t0;

        if (poisson_timer.elapsed() - ti > DELAY)
            glutPostRedisplay();

    }
    else {
        glutIdleFunc(NULL);
        glutPostRedisplay();
    }

}

#endif
#endif

int
main(int argc, char* argv[])
{

    // Declare the supported options.
    po::options_description desc("Allowed options");

    desc.add_options()
#if defined(OPENGL)
#if NDIM == 2 || NDIM == 3
        ("window,w", po::value<std::string>(), "window size (x,y)")
#endif
#endif
        ("file,f", po::value<std::string>(), "output file")
        ("radius,r", po::value<float>(&rad)->default_value(0.0f), "distribution radius")
        ("number,n", po::value<uint32_t>(&num)->default_value(0UL), "number of samples")
        ("periodic,p", "periodic boundary conditions")
        ("help", "produce help message");

    po::variables_map vm;

    try
    {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    }

    catch (const boost::program_options::error& e)
    {
        std::cerr << e.what() << std::endl;
        return EXIT_FAILURE;
    }

    if (vm.count("help"))
    {
        std::cerr << "Poisson-disk sampling in " << int(NDIM) << " dimensions\n";
        std::cerr << desc << std::endl;
        return EXIT_FAILURE;
    }

    if (rad == 0.0f && num == 0UL)
    {
        std::cerr << desc << std::endl;
        std::cerr << "Need to specify either -r or -n" << std::endl;
        return EXIT_FAILURE;
    }

    if (vm.count("periodic"))
        periodic = true;

    if (vm.count("file"))
        filename = vm["file"].as<std::string>();

#if defined(OPENGL)
#if NDIM == 2 || NDIM == 3
    if (vm.count("window"))
    {
        interactive = true;
        parse_w(resx, resy, vm["window"].as<std::string>());
    }
#endif
#endif

    if (num) {

#if NDIM != 1

        double n = 1.0;

        if (!periodic) {

#if NDIM == 2
            double a = 1.0997;
            double b = -0.499906;
#elif NDIM == 3
            double a = 2.2119;
            double b = -0.35383;
#elif NDIM == 4
            double a = 4.1114;
            double b = -0.30563;
#else
            double a = 0.0;
            double b = 0.0;
#endif

            double n_prev = 0.0;

            while (fabs(n - n_prev) > 1.0e-8) {
                n_prev = n;
                n += (num - n - a*pow(n,b+1.0))/
                     (a*(b + 1.0)*pow(n,b) + 1.0);
            }

        }
        else
            n = num;

#else
        double n = num;
#endif

#if NDIM == 2
        rad = sqrtf(0.54701f/static_cast<float>(M_PI*n));
#elif NDIM == 3
        rad = powf(1.1524f/(4.0f*static_cast<float>(M_PI*n)),1.0f/3.0f);
#elif NDIM == 4
        rad = powf(0.51981f/static_cast<float>(M_PI*M_PI*n),0.25f);
#else
        rad = 0.3684f/static_cast<float>(n);
#endif

    }

    Sample<NDIM>::setRadius(2.0f*rad);

    Counter<NDIM>::init();
    Interval::seed(static_cast<uint32_t>(time(NULL)));
    BigNum<NDIM>::seed(static_cast<uint32_t>(time(NULL)));

    gen = new Generator<NDIM>(periodic);

#if defined(OPENGL)
#  if NDIM == 2 || NDIM == 3

    if (interactive) {

        glutInit(&argc,argv);
        glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
        glutInitWindowSize(resx,resy);
        glutInitWindowPosition(100,100);
        glutCreateWindow("Poisson Disc Sampling");

        glPointSize(4.0f);
        glClearColor(1.0f,1.0f,1.0f,0.0f);

#    if NDIM == 2

        dlist = glGenLists(2);
        flist = dlist + 1UL;
        GLUquadricObj* disk = gluNewQuadric();
        gluQuadricDrawStyle(disk,GLU_SILHOUETTE);
        glNewList(dlist,GL_COMPILE);
        gluDisk(disk,0.0,rad,2000,1);
        glEndList();
        gluQuadricDrawStyle(disk,GLU_FILL);
        glNewList(flist,GL_COMPILE);
        gluDisk(disk,0.0,2.0*rad,2000,1);
        glEndList();

#    elif NDIM == 3

        clip_angles(ctheta,cphi);
        camera_angles(mtheta,mphi);

        GLfloat back_amb[] = { 3.0f, 0.0f, 0.0f };
        GLfloat back_dif[] = { 0.0f, 0.0f, 0.0f };

        slist = glGenLists(1);
        glNewList(slist,GL_COMPILE);
        glutSolidSphere(rad,20,20);
        glEndList();
        glLineWidth(4.0f);
        glEnable(GL_DEPTH_TEST);
        glMaterialfv(GL_BACK,GL_AMBIENT,back_amb);
        glMaterialfv(GL_BACK,GL_DIFFUSE,back_dif);

#    endif

#    if NDIM == 3
        glutMouseFunc(mouse);
        glutMotionFunc(motion);
#    endif
        glutIdleFunc(iterate);
        glutDisplayFunc(display);
        glutReshapeFunc(reshape);
        glutKeyboardFunc(keyboard);

        glutMainLoop();

    }

#  endif
#endif

    double t0 = poisson_timer.elapsed();

    while (gen->iterate());

    gen->output(poisson_timer.elapsed() - t0,filename);

    return EXIT_SUCCESS;

}
