// $Id: tree.h 763 2010-01-28 23:15:33Z mag $
#ifndef TREE_H_
#define TREE_H_

#include <stack>
#include <vector>
#include <cassert>
#include <boost/pool/pool.hpp>

#include "cube.h"
#include "tuple.h"
#include "bitset.h"
#include "sample.h"
#include "bignum.h"
#include "interval.h"

using namespace boost;

template <uint8_t N>
class Tree
{

    struct node
    {

        static const uint8_t M = 1 << N;
        static const uint8_t B = (M >= 8) ? M >> 3 : 1;

        node* _n;

        node();
        node(node* n);
        virtual ~node() {}

        virtual BigNum<N> volume(uint8_t l) const = 0;
        virtual bool valid(uint8_t l) const = 0;
        virtual bool leaf() const = 0;

    };

    struct inner_node : public node
    {

        static pool<> mempool;
        static uint32_t allocs;

        node* _c;
        BigNum<N> _v;
        bitset<node::B> _p;

        inner_node();
        inner_node(const bitset<node::B>& p,
                   const BigNum<N>& v,
                   node* n, node* c);
        virtual ~inner_node() {}

        void prune();

        void push(stack<const inner_node*>& s) const;

        BigNum<N> update(const Cube<N>& c,
                         uint32_t s,
                         uint8_t l);
        BigNum<N> generate(const Cube<N>& c,
                           Sample<N>& sample,
                           uint8_t l);

        virtual BigNum<N> volume(uint8_t l) const { return _v; }
        virtual bool valid(uint8_t l) const;
        virtual bool leaf() const { return false; }

        void* operator new (size_t);
        void operator delete (void* b);

    };

    struct leaf_node : public node
    {

        static BigNum<N> vol[];
        static const uint8_t LMAX = 24;

        static pool<> mempool;
        static uint32_t allocs;
        static uint32_t terminals;

        struct sample
        {

            static pool<> mempool;
            static uint32_t allocs;

            uint32_t _i;
            const sample* _n;

            sample(uint32_t i,
                   const sample* n);

            void* operator new (size_t);
            void operator delete (void* b);

        };

        const sample* _s;

        leaf_node();
        leaf_node(const sample* s,
                  leaf_node* n);
        virtual ~leaf_node();

        void append(uint32_t i);

        BigNum<N> subdivide(const Cube<N>& c,
                            uint8_t l,
                            bitset<node::B>& p,
                            leaf_node*& n);

        BigNum<N> update(const Cube<N>& c,
                         uint32_t s,
                         uint8_t l,
                         bitset<node::B>& p,
                         leaf_node*& n);
        BigNum<N> generate(const Cube<N>& c,
                           Sample<N>& sample,
                           uint8_t l,
                           bitset<node::B>& p,
                           leaf_node*& n);

        bool valid(const Sample<N>& s) const;

        virtual BigNum<N> volume(uint8_t l) const { return vol[l]; }
        virtual bool valid(uint8_t) const { return true; }
        virtual bool leaf() const { return true; }

        void* operator new (size_t);
        void operator delete (void* b);

    };

    node* _r;

    Cube<N> _c;

    typedef pair< node*, Cube<N> > NodeCube;
    typedef stack<NodeCube> NodeCubeStack;

public:

    class Iterator: private NodeCubeStack
    {

        Iterator(node* n,
                 const Cube<N>& c);
        Iterator(const NodeCubeStack& s);

        void next();

        friend class Tree;

    public:

        Iterator();

        Cube<N> operator * () const;

        Iterator& operator ++ (int);

        bool operator != (const Iterator& i) const;

    };

    Tree();

    Sample<N> generate();

    bool update(uint32_t s);

    bool empty() const;

    bool valid() const;

    double volume() const;

    Iterator end() const;
    Iterator begin() const;

    uint32_t get_terminals() const;

    static void memory_stats();

};

template <uint8_t N>
boost::pool<>
Tree<N>::inner_node::mempool(sizeof(inner_node));

template <uint8_t N>
uint32_t
Tree<N>::inner_node::allocs = 0UL;

template <uint8_t N>
uint32_t
Tree<N>::leaf_node::terminals = 0UL;

template <uint8_t N>
boost::pool<>
Tree<N>::leaf_node::mempool(sizeof(leaf_node));

template <uint8_t N>
uint32_t
Tree<N>::leaf_node::allocs = 0UL;

template <uint8_t N>
boost::pool<>
Tree<N>::leaf_node::sample::mempool(sizeof(sample));

template <uint8_t N>
uint32_t
Tree<N>::leaf_node::sample::allocs = 0UL;

template <uint8_t N>
BigNum<N>
Tree<N>::leaf_node::vol[] =
{ BigNum<N>(LMAX -  0), BigNum<N>(LMAX -  1), BigNum<N>(LMAX -  2), BigNum<N>(LMAX -  3),
  BigNum<N>(LMAX -  4), BigNum<N>(LMAX -  5), BigNum<N>(LMAX -  6), BigNum<N>(LMAX -  7),
  BigNum<N>(LMAX -  8), BigNum<N>(LMAX -  9), BigNum<N>(LMAX - 10), BigNum<N>(LMAX - 11),
  BigNum<N>(LMAX - 12), BigNum<N>(LMAX - 13), BigNum<N>(LMAX - 14), BigNum<N>(LMAX - 15),
  BigNum<N>(LMAX - 16), BigNum<N>(LMAX - 17), BigNum<N>(LMAX - 18), BigNum<N>(LMAX - 19),
  BigNum<N>(LMAX - 20), BigNum<N>(LMAX - 21), BigNum<N>(LMAX - 22), BigNum<N>(LMAX - 23),
  BigNum<N>(LMAX - 24) };

template <uint8_t N>
inline bool
Tree<N>::empty() const
{ return !_r; }

template <uint8_t N>
inline typename Tree<N>::Iterator
Tree<N>::begin() const
{ return Iterator(_r,_c); }

template <uint8_t N>
inline typename Tree<N>::Iterator
Tree<N>::end() const
{ return Iterator(); }

template <uint8_t N>
Tree<N>::node::node() : _n(0) {}

template <uint8_t N>
Tree<N>::inner_node::inner_node() : _c(0) {}

template <uint8_t N>
Tree<N>::leaf_node::leaf_node() : _s(0) {}

template <uint8_t N>
inline
Tree<N>::node::node(node* n) : _n(n) {}

template <uint8_t N>
inline
Tree<N>::inner_node::inner_node(const bitset<node::B>& p,
                                const BigNum<N>& v,
                                node* n, node* c) :
    node(n), _c(c), _v(v), _p(p) {}

template <uint8_t N>
inline
Tree<N>::leaf_node::leaf_node(const sample* s,
                              leaf_node* n) :
    node(n), _s(s) {}

template <uint8_t N>
inline void
Tree<N>::leaf_node::append(uint32_t i)
{ _s = new sample(i,_s); }

template <uint8_t N>
void
Tree<N>::inner_node::push(stack<const inner_node*>& s) const
{

    const node* c = _c;

    while (c)
    {

        if (c->leaf())
        {
            const leaf_node* l = reinterpret_cast<const leaf_node*>(c);
            c = l->_n;
            delete l;
        }
        else
        {
            const inner_node* n = reinterpret_cast<const inner_node*>(c);
            s.push(n);
            c = n->_n;
        }

    }

}

template <uint8_t N>
inline
Tree<N>::Iterator::Iterator() {}

template <uint8_t N>
inline
Tree<N>::Iterator::Iterator(node* n,
                            const Cube<N>& c)
{ if (n) { NodeCubeStack::push(NodeCube(n,c)); next(); } }

template <uint8_t N>
inline
Tree<N>::Iterator::Iterator(const NodeCubeStack& s) :
    NodeCubeStack(s) {}

template <uint8_t N>
inline Cube<N>
Tree<N>::Iterator::operator * () const
{ return !NodeCubeStack::empty() ? NodeCubeStack::top().second : Cube<N>(); }

template <uint8_t N>
inline typename Tree<N>::Iterator&
Tree<N>::Iterator::operator ++ (int)
{ if (!NodeCubeStack::empty()) { NodeCubeStack::pop(); next(); }  return *this; }

template <uint8_t N>
inline bool
Tree<N>::Iterator::operator != (const Iterator& i) const
{ const NodeCubeStack& it = reinterpret_cast<const NodeCubeStack&>(i);
  return it != *this; }

template <uint8_t N>
void
Tree<N>::Iterator::next()
{

    while (!NodeCubeStack::empty())
    {

        NodeCube nc(NodeCubeStack::top());

        if (nc.first->leaf())
            return;

        NodeCubeStack::pop();

        uint8_t i = 0;
        inner_node* n = reinterpret_cast<inner_node*>(nc.first);
        node* c = n->_c;

        while (i < node::M && !n->_p.test(i)) ++i;

        while (c)
        {

            push(NodeCube(c,nc.second.child(i)));

            c = c->_n;
            ++i;

            while (i < node::M && !n->_p.test(i)) ++i;

        }

    }

}

template <uint8_t N>
inline
Tree<N>::leaf_node::sample::sample(uint32_t i,
                                   const sample* n) :
    _i(i), _n(n) {}

template <uint8_t N>
inline
Tree<N>::Tree() :
    _r(new leaf_node()), _c(Tuple<N,Interval>(Interval())) {}

template <uint8_t N>
inline double
Tree<N>::volume() const
{

    if (_r == 0)
        return 1.0;

    if (_r->leaf())
        return 0.0;

    inner_node* n = reinterpret_cast<inner_node*>(_r);

    return 1.0 - static_cast<double>(n->_v);

}

template <uint8_t N>
Sample<N>
Tree<N>::generate()
{

    Sample<N> sample;

    if (_r)
    {

        if (_r->leaf())
        {

            leaf_node* l = reinterpret_cast<leaf_node*>(_r);

            Sample<N> s(_c.randomise());

            if (!l->valid(s))
            {

                bitset<node::B> p;
                leaf_node* n = 0;

                BigNum<N> lv(l->volume(0));
                BigNum<N> dv(l->subdivide(_c, 1, p, n));
            
                assert(!lv || dv <= lv);

                if (lv && dv == lv)
                {

                    assert(n == 0);

                    delete l;
                    _r = 0;

                }
                else
                {

                    lv -= dv;

                    _r = new inner_node(p, lv, 0, n);

                    delete l;

                }

            }
            else
                sample = s;
 
        }
        else
        {

            inner_node* n = reinterpret_cast<inner_node*>(_r);

            BigNum<N> dv(n->generate(_c, sample, 0));

            assert(!n->_v || dv <= n->_v);

            if (n->_v && dv == n->_v)
            {
                delete n;
                _r = 0;
            }
            else
                n->_v -= dv;

        }

    }

    return sample;

}

template <uint8_t N>
BigNum<N>
Tree<N>::inner_node::generate(const Cube<N>& c, Sample<N>& sample, uint8_t l)
{


    l += 1;

    assert(_c);

    uint8_t i = 0;

    node* c_prev = 0;
    node* c_curr = _c;

    while (i < node::M && !_p.test(i)) ++i;

    BigNum<N> vc(c_curr->volume(l));
    BigNum<N> pv(this->_v.random());

    while (vc && vc < pv)
    {

        ++i;
        c_prev = c_curr;
        c_curr = c_curr->_n;
        assert(c_curr);

        vc += c_curr->volume(l);

        while (i < node::M && !_p.test(i)) ++i;

    }

    if (c_curr->leaf())
    {

        bitset<node::B> p;
        leaf_node* n = 0;
        leaf_node* l_curr = reinterpret_cast<leaf_node*>(c_curr);

        BigNum<N> lv(l_curr->volume(l));
        BigNum<N> dv(l_curr->generate(c.child(i), sample, l, p, n));

        assert(dv <= lv);

        if (dv == lv)
        {

            assert(n == 0);

            _p.reset(i);
            c_curr = l_curr->_n;

            if (c_prev)
                c_prev->_n = c_curr;
            else
                _c = c_curr;

            delete l_curr;

        }
        else
        {

            lv -= dv;

            if (n)
            {

                c_curr = new inner_node(p, lv, l_curr->_n, n);

                if (c_prev)
                    c_prev->_n = c_curr;
                else
                    _c = c_curr;

                delete l_curr;

            }

        }

        return dv;

    }

    inner_node* n_curr = reinterpret_cast<inner_node*>(c_curr);

    BigNum<N> dv(n_curr->generate(c.child(i), sample, l));

    if (dv == n_curr->_v)
    {

        _p.reset(i);
        c_curr = n_curr->_n;

        if (c_prev)
            c_prev->_n = c_curr;
        else
            _c = c_curr;

        delete n_curr;

    }
    else
        n_curr->_v -= dv;

    return dv;

}

template <uint8_t N>
BigNum<N>
Tree<N>::leaf_node::generate(const Cube<N>& c, Sample<N>& sample, uint8_t l, bitset<node::B>& p, leaf_node*& n)
{

    Sample<N> s(c.randomise());

    if (!valid(s))
    {

        if (l < LMAX)
            return subdivide(c, l + 1, p, n);

        ++terminals;

        return this->volume(l);

    }
    else
    {
        sample = s;
        return BigNum<N>();
    }

}

template <uint8_t N>
void
Tree<N>::inner_node::prune()
{

    assert(_c);

    stack<const inner_node*> nstack;

    push(nstack);

    while (!nstack.empty()) {

        const inner_node* n = nstack.top();

        nstack.pop();

        n->push(nstack);

        delete n;

    }

    _c = 0;

}

template <uint8_t N>
bool
Tree<N>::leaf_node::valid(const Sample<N>& s) const
{

    if (_s)
    {

        const sample* ns = _s;

        while (ns)
        {

            if (Sample<N>::getSample(ns->_i).contains(s))
                return false;

            ns = ns->_n;

        }

    }

    return true;

}

template <uint8_t N>
bool
Tree<N>::update(uint32_t s)
{

    typename Sample<N>::Status status =
        Sample<N>::getSample(s).intersects(_c);

    if (status == Sample<N>::Out)
        return false;

    if (_r)
    {

        if (_r->leaf())
        {

            leaf_node* l = reinterpret_cast<leaf_node*>(_r);

            if (status == Sample<N>::In)
            {
                delete l;
                _r = 0;
            }
            else if (status == Sample<N>::Over)
            {

                l->append(s);

                bitset<node::B> p;
                leaf_node* n = 0;

                BigNum<N> lv(l->volume(0));
                BigNum<N> dv(l->subdivide(_c, 1, p, n));

                assert(!lv || dv <= lv);

                if (lv && dv == lv)
                {

                    assert(n == 0);

                    delete l;
                    _r = 0;

                }
                else
                {

                    lv -= dv;

                    _r = new inner_node(p, lv, 0, n);

                    delete l;

                }

            }

        }
        else
        {

            inner_node* n = reinterpret_cast<inner_node*>(_r);

            BigNum<N> dv(n->update(_c, s, 0));

            assert(!n->_v || dv <= n->_v);

            if (n->_v && dv == n->_v)
            {
                delete n;
                _r = 0;
            }
            else
                n->_v -= dv;

        }

    }

    return true;

}

template <uint8_t N>
BigNum<N>
Tree<N>::inner_node::update(const Cube<N>& c, uint32_t s, uint8_t l)
{

    assert(_c);

    typename Sample<N>::Status status =
        Sample<N>::getSample(s).intersects(c);

    if (status == Sample<N>::Out)
        return BigNum<N>();

    if (status == Sample<N>::In)
    {
        prune();
        return this->_v;
    }

    l += 1;

    BigNum<N> dv;

    uint8_t i = 0;

    node* c_prev = 0;
    node* c_curr = _c;


    while (i < node::M && !_p.test(i)) ++i;

    while (c_curr)
    {

        Cube<N> ci(c.child(i));

        if (c_curr->leaf())
        {

            bitset<node::B> p;
            leaf_node* n = 0;
            leaf_node* l_curr = reinterpret_cast<leaf_node*>(c_curr);

            BigNum<N> lv(l_curr->volume(l));
            BigNum<N> dvl(l_curr->update(ci, s, l, p, n));

            assert(dvl <= lv);

            if (dvl == lv)
            {

                assert(n == 0);

                _p.reset(i);
                c_curr = l_curr->_n;

                if (c_prev)
                    c_prev->_n = c_curr;
                else
                    _c = c_curr;

                delete l_curr;

            }
            else
            {

                lv -= dvl;

                if (n)
                {

                    c_curr = new inner_node(p, lv, l_curr->_n, n);

                    if (c_prev)
                        c_prev->_n = c_curr;
                    else
                        _c = c_curr;

                    delete l_curr;

                }

                c_prev = c_curr;
                c_curr = c_curr->_n;

            }

            dv += dvl;

        }
        else
        {

            inner_node* n_curr = reinterpret_cast<inner_node*>(c_curr);

            BigNum<N> dvc(n_curr->update(ci, s, l));

            assert(dvc <= n_curr->_v);

            if (dvc == n_curr->_v)
            {

                _p.reset(i);
                c_curr = n_curr->_n;

                if (c_prev)
                    c_prev->_n = c_curr;
                else
                    _c = c_curr;

                delete n_curr;

            }
            else
            {
                n_curr->_v -= dvc;
                c_prev = n_curr;
                c_curr = n_curr->_n;
            }

            dv += dvc;

        }

        ++i;

        while (i < node::M && !_p.test(i)) ++i;

    }

    return dv;

}

template <uint8_t N>
BigNum<N>
Tree<N>::leaf_node::update(const Cube<N>& c, uint32_t s, uint8_t l, bitset<node::B>& p, leaf_node*& n)
{

    typename Sample<N>::Status status =
        Sample<N>::getSample(s).intersects(c);

    if (status == Sample<N>::Out)
        return BigNum<N>();

    if (status == Sample<N>::In)
        return this->volume(l);

    append(s);

    return subdivide(c, l + 1, p, n);

}

template <uint8_t N>
inline bool
Tree<N>::valid() const
{

    if (_r == 0)
        return inner_node::allocs == 0UL &&
               leaf_node::allocs == 0UL &&
               leaf_node::sample::allocs == 0UL;

    if (_r->leaf())
        return _r->volume(0) == BigNum<N>();

    return _r->valid(0);

}

template <uint8_t N>
bool
Tree<N>::inner_node::valid(uint8_t l) const
{

    BigNum<N> v;
    uint8_t i = 0;
    const node* c = _c;

    l += 1;

    while (c)
    {
        v += c->volume(l);
        c = c->_n;
        ++i;
    }

    if (this->_v != v || _p.count() != i)
        return false;

    c = _c;

    while (c)
    {
        if (!c->valid(l))
            return false;
        c = c->_n;
    }

    return true;

}

template <uint8_t N>
Tree<N>::leaf_node::~leaf_node()
{

    while (_s)
    {

        const sample* s = _s->_n;

        delete _s;

        _s = s;

    }

}

template <uint8_t N>
BigNum<N>
Tree<N>::leaf_node::subdivide(const Cube<N>& c, uint8_t l, bitset<node::B>& p, leaf_node*& n)
{

    Cube<N> cube[node::M];

    c.subdivide(cube);

    BigNum<N> dv;
    BigNum<N> dvc(LMAX - l);

    for (uint8_t i = node::M; i > 0; i--) {

        uint8_t j = i - 1;

        const sample* s = _s;
        const sample* sc = 0;

        while (s) {

            typename Sample<N>::Status status = 
                Sample<N>::getSample(s->_i).intersects(cube[j]);

            if (status == Sample<N>::In)
                break;

            if (status == Sample<N>::Over)
                sc = new sample(s->_i,sc);

            s = s->_n;

        }

        if (s) {

            while (sc) {

                const sample* s = sc->_n;

                delete sc;

                sc = s;

            }

            dv += dvc;

        }
        else
        {
            n = new leaf_node(sc,n);
            p.set(j);
        }

    }

    while (_s) {

        const sample* s = _s->_n;

        delete _s;

        _s = s;

    }

    return dv;

}

template <uint8_t N>
void*
Tree<N>::inner_node::operator new (size_t)
{

    void* p = 0;

    while (p == 0)
    {

        p = mempool.malloc();

        if (p == 0)
        {

            pool<>::size_type next_size = mempool.get_next_size() >> 1;

            if (next_size < sizeof(node))
            {
                cerr << "\nTree<N>::inner_node: Unable to allocate more memory. Exiting...\n";
                exit(EXIT_FAILURE);
            }
            else
                mempool.set_next_size(next_size);

        }

    }

    ++allocs;

    return p;

}

template <uint8_t N>
void*
Tree<N>::leaf_node::operator new (size_t)
{

    void* p = 0;

    while (p == 0)
    {

        p = mempool.malloc();

        if (p == 0)
        {

            pool<>::size_type next_size = mempool.get_next_size() >> 1;

            if (next_size < sizeof(leaf_node))
            {
                cerr << "\nTree<N>::leaf_node: Unable to allocate more memory. Exiting...\n";
                exit(EXIT_FAILURE);
            }
            else
                mempool.set_next_size(next_size);

        }

    }

    ++allocs;

    return p;

}

template <uint8_t N>
inline void
Tree<N>::inner_node::operator delete (void* b)
{ mempool.free(b); --allocs; }

template <uint8_t N>
inline void
Tree<N>::leaf_node::operator delete (void* b)
{ mempool.free(b); --allocs; }

template <uint8_t N>
inline void*
Tree<N>::leaf_node::sample::operator new (size_t)
{

    void* p = 0;

    while (p == 0)
    {

        p = mempool.malloc();

        if (p == 0)
        {

            pool<>::size_type next_size = mempool.get_next_size() >> 1;

            if (next_size < sizeof(typename leaf_node::sample))
            {
                cerr << "\nTree<N>::leaf_node::sample: Unable to allocate more memory. Exiting...\n";
                exit(EXIT_FAILURE);
            }
            else
                mempool.set_next_size(next_size);

        }

    }

    ++allocs;

    return p;

}

template <uint8_t N>
inline void
Tree<N>::leaf_node::sample::operator delete (void* b)
{ mempool.free(b); --allocs; }

template <uint8_t N>
inline uint32_t
Tree<N>::get_terminals() const
{ return leaf_node::terminals; }

template <uint8_t N>
inline void
Tree<N>::memory_stats()
{ cout << inner_node::allocs << ' ' << leaf_node::allocs << ' ' << leaf_node::sample::allocs << endl; }

#endif /*TREE_H_*/
